import pandas as pd

from ..base import VannaBase


class MockVectorDB(VannaBase):
    def __init__(self, config=None):
        pass

    def _get_id(self, value: str, **kwargs) -> str:
        # Hash the value and return the ID
        return str(hash(value))

    def add_ddl(self, ddl: str, **kwargs) -> str:
        return self._get_id(ddl)

    def add_documentation(self, doc: str, **kwargs) -> str:
        return self._get_id(doc)

    def add_question_sql(self, question: str, sql: str, **kwargs) -> str:
        return self._get_id(question)

    def get_related_ddl(self, question: str, **kwargs) -> list:
        return []

    def get_related_documentation(self, question: str, **kwargs) -> list:
        return []

    def get_similar_question_sql(self, question: str, **kwargs) -> list:
        return []

    def get_training_data(self, **kwargs) -> pd.DataFrame:
        return pd.DataFrame({'id': {0: '19546-ddl',
  1: '91597-sql',
  2: '133976-sql',
  3: '59851-doc',
  4: '73046-sql'},
 'training_data_type': {0: 'ddl',
  1: 'sql',
  2: 'sql',
  3: 'documentation',
  4: 'sql'},
 'question': {0: None,
  1: 'What are the top selling genres?',
  2: 'What are the low 7 artists by sales?',
  3: None,
  4: 'What is the total sales for each customer?'},
 'content': {0: 'CREATE TABLE [Invoice]\n(\n    [InvoiceId] INTEGER  NOT NULL,\n    [CustomerId] INTEGER  NOT NULL,\n    [InvoiceDate] DATETIME  NOT NULL,\n    [BillingAddress] NVARCHAR(70),\n    [BillingCity] NVARCHAR(40),\n    [BillingState] NVARCHAR(40),\n    [BillingCountry] NVARCHAR(40),\n    [BillingPostalCode] NVARCHAR(10),\n    [Total] NUMERIC(10,2)  NOT NULL,\n    CONSTRAINT [PK_Invoice] PRIMARY KEY  ([InvoiceId]),\n    FOREIGN KEY ([CustomerId]) REFERENCES [Customer] ([CustomerId]) \n\t\tON DELETE NO ACTION ON UPDATE NO ACTION\n)',
  1: 'SELECT g.Name AS Genre, SUM(il.Quantity) AS TotalSales\nFROM Genre g\nJOIN Track t ON g.GenreId = t.GenreId\nJOIN InvoiceLine il ON t.TrackId = il.TrackId\nGROUP BY g.GenreId, g.Name\nORDER BY TotalSales DESC;',
  2: 'SELECT a.ArtistId, a.Name, SUM(il.Quantity) AS TotalSales\nFROM Artist a\nINNER JOIN Album al ON a.ArtistId = al.ArtistId\nINNER JOIN Track t ON al.AlbumId = t.AlbumId\nINNER JOIN InvoiceLine il ON t.TrackId = il.TrackId\nGROUP BY a.ArtistId, a.Name\nORDER BY TotalSales ASC\nLIMIT 7;',
  3: 'This is a SQLite database. For dates rememeber to use SQLite syntax.',
  4: 'SELECT c.CustomerId, c.FirstName, c.LastName, SUM(i.Total) AS TotalSales\nFROM Customer c\nJOIN Invoice i ON c.CustomerId = i.CustomerId\nGROUP BY c.CustomerId, c.FirstName, c.LastName;'}})

    def remove_training_data(id: str, **kwargs) -> bool:
        return True



import uuid
from typing import List

import pandas as pd
from pymilvus import DataType, MilvusClient, model

from ..base import VannaBase

# Setting the URI as a local file, e.g.`./milvus.db`,
# is the most convenient method, as it automatically utilizes Milvus Lite
# to store all data in this file.
#
# If you have large scale of data such as more than a million docs, we
# recommend setting up a more performant Milvus server on docker or kubernetes.
# When using this setup, please use the server URI,
# e.g.`http://localhost:19530`, as your URI.

DEFAULT_MILVUS_URI = "./milvus.db"
# DEFAULT_MILVUS_URI = "http://localhost:19530"

MAX_LIMIT_SIZE = 10_000


class Milvus_VectorStore(VannaBase):
    """
    Vectorstore implementation using Milvus - https://milvus.io/docs/quickstart.md

    Args:
        - config (dict, optional): Dictionary of `Milvus_VectorStore config` options. Defaults to `None`.
            - milvus_client: A `pymilvus.MilvusClient` instance.
            - embedding_function:
                A `milvus_model.base.BaseEmbeddingFunction` instance. Defaults to `DefaultEmbeddingFunction()`.
                For more models, please refer to:
                https://milvus.io/docs/embeddings.md
    """
    def __init__(self, config=None):
        VannaBase.__init__(self, config=config)

        if "milvus_client" in config:
            self.milvus_client = config["milvus_client"]
        else:
            self.milvus_client = MilvusClient(uri=DEFAULT_MILVUS_URI)

        if "embedding_function" in config:
            self.embedding_function = config.get("embedding_function")
        else:
            self.embedding_function = model.DefaultEmbeddingFunction()
        self._embedding_dim = self.embedding_function.encode_documents(["foo"])[0].shape[0]
        self._create_collections()
        self.n_results = config.get("n_results", 10)

    def _create_collections(self):
        self._create_sql_collection("vannasql")
        self._create_ddl_collection("vannaddl")
        self._create_doc_collection("vannadoc")


    def generate_embedding(self, data: str, **kwargs) -> List[float]:
        return self.embedding_function.encode_documents(data).tolist()


    def _create_sql_collection(self, name: str):
        if not self.milvus_client.has_collection(collection_name=name):
            vannasql_schema = MilvusClient.create_schema(
                auto_id=False,
                enable_dynamic_field=False,
            )
            vannasql_schema.add_field(field_name="id", datatype=DataType.VARCHAR, max_length=65535, is_primary=True)
            vannasql_schema.add_field(field_name="text", datatype=DataType.VARCHAR, max_length=65535)
            vannasql_schema.add_field(field_name="sql", datatype=DataType.VARCHAR, max_length=65535)
            vannasql_schema.add_field(field_name="vector", datatype=DataType.FLOAT_VECTOR, dim=self._embedding_dim)

            vannasql_index_params = self.milvus_client.prepare_index_params()
            vannasql_index_params.add_index(
                field_name="vector",
                index_name="vector",
                index_type="AUTOINDEX",
                metric_type="L2",
            )
            self.milvus_client.create_collection(
                collection_name=name,
                schema=vannasql_schema,
                index_params=vannasql_index_params,
                consistency_level="Strong"
            )

    def _create_ddl_collection(self, name: str):
        if not self.milvus_client.has_collection(collection_name=name):
            vannaddl_schema = MilvusClient.create_schema(
                auto_id=False,
                enable_dynamic_field=False,
            )
            vannaddl_schema.add_field(field_name="id", datatype=DataType.VARCHAR, max_length=65535, is_primary=True)
            vannaddl_schema.add_field(field_name="ddl", datatype=DataType.VARCHAR, max_length=65535)
            vannaddl_schema.add_field(field_name="vector", datatype=DataType.FLOAT_VECTOR, dim=self._embedding_dim)

            vannaddl_index_params = self.milvus_client.prepare_index_params()
            vannaddl_index_params.add_index(
                field_name="vector",
                index_name="vector",
                index_type="AUTOINDEX",
                metric_type="L2",
            )
            self.milvus_client.create_collection(
                collection_name=name,
                schema=vannaddl_schema,
                index_params=vannaddl_index_params,
                consistency_level="Strong"
            )

    def _create_doc_collection(self, name: str):
        if not self.milvus_client.has_collection(collection_name=name):
            vannadoc_schema = MilvusClient.create_schema(
                auto_id=False,
                enable_dynamic_field=False,
            )
            vannadoc_schema.add_field(field_name="id", datatype=DataType.VARCHAR, max_length=65535, is_primary=True)
            vannadoc_schema.add_field(field_name="doc", datatype=DataType.VARCHAR, max_length=65535)
            vannadoc_schema.add_field(field_name="vector", datatype=DataType.FLOAT_VECTOR, dim=self._embedding_dim)

            vannadoc_index_params = self.milvus_client.prepare_index_params()
            vannadoc_index_params.add_index(
                field_name="vector",
                index_name="vector",
                index_type="AUTOINDEX",
                metric_type="L2",
            )
            self.milvus_client.create_collection(
                collection_name=name,
                schema=vannadoc_schema,
                index_params=vannadoc_index_params,
                consistency_level="Strong"
            )

    def add_question_sql(self, question: str, sql: str, **kwargs) -> str:
        if len(question) == 0 or len(sql) == 0:
            raise Exception("pair of question and sql can not be null")
        _id = str(uuid.uuid4()) + "-sql"
        embedding = self.embedding_function.encode_documents([question])[0]
        self.milvus_client.insert(
            collection_name="vannasql",
            data={
                "id": _id,
                "text": question,
                "sql": sql,
                "vector": embedding
            }
        )
        return _id

    def add_ddl(self, ddl: str, **kwargs) -> str:
        if len(ddl) == 0:
            raise Exception("ddl can not be null")
        _id = str(uuid.uuid4()) + "-ddl"
        embedding = self.embedding_function.encode_documents([ddl])[0]
        self.milvus_client.insert(
            collection_name="vannaddl",
            data={
                "id": _id,
                "ddl": ddl,
                "vector": embedding
            }
        )
        return _id

    def add_documentation(self, documentation: str, **kwargs) -> str:
        if len(documentation) == 0:
            raise Exception("documentation can not be null")
        _id = str(uuid.uuid4()) + "-doc"
        embedding = self.embedding_function.encode_documents([documentation])[0]
        self.milvus_client.insert(
            collection_name="vannadoc",
            data={
                "id": _id,
                "doc": documentation,
                "vector": embedding
            }
        )
        return _id

    def get_training_data(self, **kwargs) -> pd.DataFrame:
        sql_data = self.milvus_client.query(
            collection_name="vannasql",
            output_fields=["*"],
            limit=MAX_LIMIT_SIZE,
        )
        df = pd.DataFrame()
        df_sql = pd.DataFrame(
            {
                "id": [doc["id"] for doc in sql_data],
                "question": [doc["text"] for doc in sql_data],
                "content": [doc["sql"] for doc in sql_data],
            }
        )
        df = pd.concat([df, df_sql])

        ddl_data = self.milvus_client.query(
            collection_name="vannaddl",
            output_fields=["*"],
            limit=MAX_LIMIT_SIZE,
        )

        df_ddl = pd.DataFrame(
            {
                "id": [doc["id"] for doc in ddl_data],
                "question": [None for doc in ddl_data],
                "content": [doc["ddl"] for doc in ddl_data],
            }
        )
        df = pd.concat([df, df_ddl])

        doc_data = self.milvus_client.query(
            collection_name="vannadoc",
            output_fields=["*"],
            limit=MAX_LIMIT_SIZE,
        )

        df_doc = pd.DataFrame(
            {
                "id": [doc["id"] for doc in doc_data],
                "question": [None for doc in doc_data],
                "content": [doc["doc"] for doc in doc_data],
            }
        )
        df = pd.concat([df, df_doc])
        return df

    def get_similar_question_sql(self, question: str, **kwargs) -> list:
        search_params = {
            "metric_type": "L2",
            "params": {"nprobe": 128},
        }
        embeddings = self.embedding_function.encode_queries([question])
        res = self.milvus_client.search(
            collection_name="vannasql",
            anns_field="vector",
            data=embeddings,
            limit=self.n_results,
            output_fields=["text", "sql"],
            search_params=search_params
        )
        res = res[0]

        list_sql = []
        for doc in res:
            dict = {}
            dict["question"] = doc["entity"]["text"]
            dict["sql"] = doc["entity"]["sql"]
            list_sql.append(dict)
        return list_sql

    def get_related_ddl(self, question: str, **kwargs) -> list:
        search_params = {
            "metric_type": "L2",
            "params": {"nprobe": 128},
        }
        embeddings = self.embedding_function.encode_queries([question])
        res = self.milvus_client.search(
            collection_name="vannaddl",
            anns_field="vector",
            data=embeddings,
            limit=self.n_results,
            output_fields=["ddl"],
            search_params=search_params
        )
        res = res[0]

        list_ddl = []
        for doc in res:
            list_ddl.append(doc["entity"]["ddl"])
        return list_ddl

    def get_related_documentation(self, question: str, **kwargs) -> list:
        search_params = {
            "metric_type": "L2",
            "params": {"nprobe": 128},
        }
        embeddings = self.embedding_function.encode_queries([question])
        res = self.milvus_client.search(
            collection_name="vannadoc",
            anns_field="vector",
            data=embeddings,
            limit=self.n_results,
            output_fields=["doc"],
            search_params=search_params
        )
        res = res[0]

        list_doc = []
        for doc in res:
            list_doc.append(doc["entity"]["doc"])
        return list_doc

    def remove_training_data(self, id: str, **kwargs) -> bool:
        if id.endswith("-sql"):
            self.milvus_client.delete(collection_name="vannasql", ids=[id])
            return True
        elif id.endswith("-ddl"):
            self.milvus_client.delete(collection_name="vannaddl", ids=[id])
            return True
        elif id.endswith("-doc"):
            self.milvus_client.delete(collection_name="vannadoc", ids=[id])
            return True
        else:
            return False



import weaviate
import weaviate.classes as wvc
from fastembed import TextEmbedding

from vanna.base import VannaBase


class WeaviateDatabase(VannaBase):

    def __init__(self, config=None):
        """
        Initialize the VannaEnhanced class with the provided configuration.

        :param config: Dictionary containing configuration parameters.

        params:
        weaviate_url (str): Weaviate cluster URL while using weaviate cloud,
        weaviate_api_key (str): Weaviate API key while using weaviate cloud,
        weaviate_port (num): Weaviate port while using local weaviate,
        weaviate_grpc (num): Weaviate gRPC port while using local weaviate,
        fastembed_model (str): Fastembed model name for text embeddings. BAAI/bge-small-en-v1.5 by default.

        """
        super().__init__(config=config)

        if config is None:
            raise ValueError("config is required")

        self.fastembed_model = config.get("fastembed_model", "BAAI/bge-small-en-v1.5")
        self.weaviate_api_key = config.get("weaviate_api_key")
        self.weaviate_url = config.get("weaviate_url")
        self.weaviate_port = config.get("weaviate_port")
        self.weaviate_grpc_port = config.get("weaviate_grpc", 50051)

        if not self.weaviate_api_key and not self.weaviate_port:
            raise ValueError("Add proper credentials to connect to weaviate")

        self.weaviate_client = self._initialize_weaviate_client()
        self.embeddings = TextEmbedding(model_name=self.fastembed_model)

        self.training_data_cluster = {
            "sql": "SQLTrainingDataEntry",
            "ddl": "DDLEntry",
            "doc": "DocumentationEntry"
        }

        self._create_collections_if_not_exist()

    def _create_collections_if_not_exist(self):
        properties_dict = {
            self.training_data_cluster['ddl']: [
                wvc.config.Property(name="description", data_type=wvc.config.DataType.TEXT),
            ],
            self.training_data_cluster['doc']: [
                wvc.config.Property(name="description", data_type=wvc.config.DataType.TEXT),
            ],
            self.training_data_cluster['sql']: [
                wvc.config.Property(name="sql", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(name="natural_language_question", data_type=wvc.config.DataType.TEXT),
            ]
        }

        for cluster, properties in properties_dict.items():
            if not self.weaviate_client.collections.exists(cluster):
                self.weaviate_client.collections.create(
                    name=cluster,
                    properties=properties
                )

    def _initialize_weaviate_client(self):
        if self.weaviate_api_key:
            return weaviate.connect_to_wcs(
                cluster_url=self.weaviate_url,
                auth_credentials=weaviate.auth.AuthApiKey(self.weaviate_api_key),
                additional_config=weaviate.config.AdditionalConfig(timeout=(10, 300)),
                skip_init_checks=True
            )
        else:
            return weaviate.connect_to_local(
                port=self.weaviate_port,
                grpc_port=self.weaviate_grpc_port,
                additional_config=weaviate.config.AdditionalConfig(timeout=(10, 300)),
                skip_init_checks=True
            )

    def generate_embedding(self, data: str, **kwargs):
            embedding_model = TextEmbedding(model_name=self.fastembed_model)
            embedding = next(embedding_model.embed(data))
            return embedding.tolist()


    def _insert_data(self, cluster_key: str, data_object: dict, vector: list) -> str:
        self.weaviate_client.connect()
        response = self.weaviate_client.collections.get(self.training_data_cluster[cluster_key]).data.insert(
            properties=data_object,
            vector=vector
        )
        self.weaviate_client.close()
        return response

    def add_ddl(self, ddl: str, **kwargs) -> str:
        data_object = {
            "description": ddl,
        }
        response = self._insert_data('ddl', data_object, self.generate_embedding(ddl))
        return f'{response}-ddl'

    def add_documentation(self, doc: str, **kwargs) -> str:
        data_object = {
            "description": doc,
        }
        response = self._insert_data('doc', data_object, self.generate_embedding(doc))
        return f'{response}-doc'

    def add_question_sql(self, question: str, sql: str, **kwargs) -> str:
        data_object = {
            "sql": sql,
            "natural_language_question": question,
        }
        response = self._insert_data('sql', data_object, self.generate_embedding(question))
        return f'{response}-sql'

    def _query_collection(self, cluster_key: str, vector_input: list, return_properties: list, limit: int = 3) -> list:
        self.weaviate_client.connect()
        collection = self.weaviate_client.collections.get(self.training_data_cluster[cluster_key])
        response = collection.query.near_vector(
            near_vector=vector_input,
            limit=limit,
            return_properties=return_properties
        )
        response_list = [item.properties for item in response.objects]
        self.weaviate_client.close()
        return response_list

    def get_related_ddl(self, question: str, **kwargs) -> list:
        vector_input = self.generate_embedding(question)
        response_list = self._query_collection('ddl', vector_input, ["description"])
        return [item["description"] for item in response_list]

    def get_related_documentation(self, question: str, **kwargs) -> list:
        vector_input = self.generate_embedding(question)
        response_list = self._query_collection('doc', vector_input, ["description"])
        return [item["description"] for item in response_list]

    def get_similar_question_sql(self, question: str, **kwargs) -> list:
        vector_input = self.generate_embedding(question)
        response_list = self._query_collection('sql', vector_input, ["sql", "natural_language_question"])
        return [{"question": item["natural_language_question"], "sql": item["sql"]} for item in response_list]

    def get_training_data(self, **kwargs) -> list:
        self.weaviate_client.connect()
        combined_response_list = []
        for collection_name in self.training_data_cluster.values():
            if self.weaviate_client.collections.exists(collection_name):
                collection = self.weaviate_client.collections.get(collection_name)
                response_list = [item.properties for item in collection.iterator()]
                combined_response_list.extend(response_list)
        self.weaviate_client.close()
        return combined_response_list

    def remove_training_data(self, id: str, **kwargs) -> bool:
        self.weaviate_client.connect()
        success = False
        if id.endswith("-sql"):
            id = id.replace('-sql', '')
            success = self.weaviate_client.collections.get(self.training_data_cluster['sql']).data.delete_by_id(id)
        elif id.endswith("-ddl"):
            id = id.replace('-ddl', '')
            success = self.weaviate_client.collections.get(self.training_data_cluster['ddl']).data.delete_by_id(id)
        elif id.endswith("-doc"):
            id = id.replace('-doc', '')
            success = self.weaviate_client.collections.get(self.training_data_cluster['doc']).data.delete_by_id(id)
        self.weaviate_client.close()
        return success



import re
from transformers import AutoTokenizer, AutoModelForCausalLM

from ..base import VannaBase


class Hf(VannaBase):
    def __init__(self, config=None):
        model_name_or_path = self.config.get(
            "model_name_or_path", None
        )  # e.g. meta-llama/Meta-Llama-3-8B-Instruct or local path to the model checkpoint files
        # list of quantization methods supported by transformers package: https://huggingface.co/docs/transformers/main/en/quantization/overview
        quantization_config = self.config.get("quantization_config", None)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        self.model = AutoModelForCausalLM.from_pretrained(
            model_name_or_path,
            quantization_config=quantization_config,
            device_map="auto",
        )

    def system_message(self, message: str) -> any:
        return {"role": "system", "content": message}

    def user_message(self, message: str) -> any:
        return {"role": "user", "content": message}

    def assistant_message(self, message: str) -> any:
        return {"role": "assistant", "content": message}

    def extract_sql_query(self, text):
        """
        Extracts the first SQL statement after the word 'select', ignoring case,
        matches until the first semicolon, three backticks, or the end of the string,
        and removes three backticks if they exist in the extracted string.

        Args:
        - text (str): The string to search within for an SQL statement.

        Returns:
        - str: The first SQL statement found, with three backticks removed, or an empty string if no match is found.
        """
        # Regular expression to find 'select' (ignoring case) and capture until ';', '```', or end of string
        pattern = re.compile(r"select.*?(?:;|```|$)", re.IGNORECASE | re.DOTALL)

        match = pattern.search(text)
        if match:
            # Remove three backticks from the matched string if they exist
            return match.group(0).replace("```", "")
        else:
            return text

    def generate_sql(self, question: str, **kwargs) -> str:
        # Use the super generate_sql
        sql = super().generate_sql(question, **kwargs)

        # Replace "\_" with "_"
        sql = sql.replace("\\_", "_")

        sql = sql.replace("\\", "")

        return self.extract_sql_query(sql)

    def submit_prompt(self, prompt, **kwargs) -> str:

        input_ids = self.tokenizer.apply_chat_template(
            prompt, add_generation_prompt=True, return_tensors="pt"
        ).to(self.model.device)

        outputs = self.model.generate(
            input_ids,
            max_new_tokens=512,
            eos_token_id=self.tokenizer.eos_token_id,
            do_sample=True,
            temperature=1,
            top_p=0.9,
        )
        response = outputs[0][input_ids.shape[-1] :]
        response = self.tokenizer.decode(response, skip_special_tokens=True)
        self.log(response)

        return response



from mistralai.client import MistralClient
from mistralai.models.chat_completion import ChatMessage

from ..base import VannaBase


class Mistral(VannaBase):
    def __init__(self, config=None):
        if config is None:
            raise ValueError(
                "For Mistral, config must be provided with an api_key and model"
            )

        if "api_key" not in config:
            raise ValueError("config must contain a Mistral api_key")

        if "model" not in config:
            raise ValueError("config must contain a Mistral model")

        api_key = config["api_key"]
        model = config["model"]
        self.client = MistralClient(api_key=api_key)
        self.model = model

    def system_message(self, message: str) -> any:
        return ChatMessage(role="system", content=message)

    def user_message(self, message: str) -> any:
        return ChatMessage(role="user", content=message)

    def assistant_message(self, message: str) -> any:
        return ChatMessage(role="assistant", content=message)

    def generate_sql(self, question: str, **kwargs) -> str:
        # Use the super generate_sql
        sql = super().generate_sql(question, **kwargs)

        # Replace "\_" with "_"
        sql = sql.replace("\\_", "_")

        return sql

    def submit_prompt(self, prompt, **kwargs) -> str:
        chat_response = self.client.chat(
            model=self.model,
            messages=prompt,
        )

        return chat_response.choices[0].message.content



from abc import ABC, abstractmethod


class VannaAdvanced(ABC):
    def __init__(self, config=None):
        self.config = config

    @abstractmethod
    def get_function(self, question: str, additional_data: dict = {}) -> dict:
        pass

    @abstractmethod
    def create_function(self, question: str, sql: str, plotly_code: str, **kwargs) -> dict:
        pass

    @abstractmethod
    def update_function(self, old_function_name: str, updated_function: dict) -> bool:
        pass

    @abstractmethod
    def delete_function(self, function_name: str) -> bool:
        pass

    @abstractmethod
    def get_all_functions(self) -> list:
        pass



from abc import ABC, abstractmethod

import flask


class AuthInterface(ABC):
    @abstractmethod
    def get_user(self, flask_request) -> any:
        pass

    @abstractmethod
    def is_logged_in(self, user: any) -> bool:
        pass

    @abstractmethod
    def override_config_for_user(self, user: any, config: dict) -> dict:
        pass

    @abstractmethod
    def login_form(self) -> str:
        pass

    @abstractmethod
    def login_handler(self, flask_request) -> str:
        pass

    @abstractmethod
    def callback_handler(self, flask_request) -> str:
        pass

    @abstractmethod
    def logout_handler(self, flask_request) -> str:
        pass

class NoAuth(AuthInterface):
    def get_user(self, flask_request) -> any:
        return {}

    def is_logged_in(self, user: any) -> bool:
        return True

    def override_config_for_user(self, user: any, config: dict) -> dict:
        return config

    def login_form(self) -> str:
        return ''

    def login_handler(self, flask_request) -> str:
        return 'No login required'

    def callback_handler(self, flask_request) -> str:
        return 'No login required'

    def logout_handler(self, flask_request) -> str:
        return 'No login required'



from typing import List

from ..base import VannaBase


class MockEmbedding(VannaBase):
    def __init__(self, config=None):
        pass

    def generate_embedding(self, data: str, **kwargs) -> List[float]:
        return [1.0, 2.0, 3.0, 4.0, 5.0]



import hashlib
import os
import re
import uuid
from typing import Union

from .exceptions import ImproperlyConfigured, ValidationError


def validate_config_path(path):
    if not os.path.exists(path):
        raise ImproperlyConfigured(
            f'No such configuration file: {path}'
        )

    if not os.path.isfile(path):
        raise ImproperlyConfigured(
            f'Config should be a file: {path}'
        )

    if not os.access(path, os.R_OK):
        raise ImproperlyConfigured(
            f'Cannot read the config file. Please grant read privileges: {path}'
        )


def sanitize_model_name(model_name):
    try:
        model_name = model_name.lower()

        # Replace spaces with a hyphen
        model_name = model_name.replace(" ", "-")

        if '-' in model_name:

            # remove double hyphones
            model_name = re.sub(r"-+", "-", model_name)
            if '_' in model_name:
                # If name contains both underscores and hyphen replace all underscores with hyphens
                model_name = re.sub(r'_', '-', model_name)

        # Remove special characters only allow underscore
        model_name = re.sub(r"[^a-zA-Z0-9-_]", "", model_name)

        # Remove hyphen or underscore if any at the last or first
        if model_name[-1] in ("-", "_"):
            model_name = model_name[:-1]
        if model_name[0] in ("-", "_"):
            model_name = model_name[1:]

        return model_name
    except Exception as e:
        raise ValidationError(e)


def deterministic_uuid(content: Union[str, bytes]) -> str:
    """Creates deterministic UUID on hash value of string or byte content.

    Args:
        content: String or byte representation of data.

    Returns:
        UUID of the content.
    """
    if isinstance(content, str):
        content_bytes = content.encode("utf-8")
    elif isinstance(content, bytes):
        content_bytes = content
    else:
        raise ValueError(f"Content type {type(content)} not supported !")

    hash_object = hashlib.sha256(content_bytes)
    hash_hex = hash_object.hexdigest()
    namespace = uuid.UUID("00000000-0000-0000-0000-000000000000")
    content_uuid = str(uuid.uuid5(namespace, hash_hex))

    return content_uuid



import dataclasses
import json
from io import StringIO

import pandas as pd
import requests

from ..advanced import VannaAdvanced
from ..base import VannaBase
from ..types import (
  DataFrameJSON,
  NewOrganization,
  OrganizationList,
  Question,
  QuestionSQLPair,
  Status,
  StatusWithId,
  StringData,
  TrainingData,
)
from ..utils import sanitize_model_name


class VannaDB_VectorStore(VannaBase, VannaAdvanced):
    def __init__(self, vanna_model: str, vanna_api_key: str, config=None):
        VannaBase.__init__(self, config=config)

        self._model = vanna_model
        self._api_key = vanna_api_key

        self._endpoint = (
            "https://ask.vanna.ai/rpc"
            if config is None or "endpoint" not in config
            else config["endpoint"]
        )
        self.related_training_data = {}
        self._graphql_endpoint = "https://functionrag.com/query"
        self._graphql_headers = {
            "Content-Type": "application/json",
            "API-KEY": self._api_key,
            "NAMESPACE": self._model,
        }

    def _rpc_call(self, method, params):
        if method != "list_orgs":
            headers = {
                "Content-Type": "application/json",
                "Vanna-Key": self._api_key,
                "Vanna-Org": self._model,
            }
        else:
            headers = {
                "Content-Type": "application/json",
                "Vanna-Key": self._api_key,
                "Vanna-Org": "demo-tpc-h",
            }

        data = {
            "method": method,
            "params": [self._dataclass_to_dict(obj) for obj in params],
        }

        response = requests.post(self._endpoint, headers=headers, data=json.dumps(data))
        return response.json()

    def _dataclass_to_dict(self, obj):
        return dataclasses.asdict(obj)

    def get_all_functions(self) -> list:
        query = """
            {
                get_all_sql_functions {
                    function_name
                    description
                    post_processing_code_template
                    arguments {
                        name
                        description
                        general_type
                        is_user_editable
                        available_values
                    }
                    sql_template
                }
            }
        """

        response = requests.post(self._graphql_endpoint, headers=self._graphql_headers, json={'query': query})
        response_json = response.json()
        if response.status_code == 200 and 'data' in response_json and 'get_all_sql_functions' in response_json['data']:
            self.log(response_json['data']['get_all_sql_functions'])
            resp = response_json['data']['get_all_sql_functions']

            print(resp)

            return resp
        else:
            raise Exception(f"Query failed to run by returning code of {response.status_code}. {response.text}")


    def get_function(self, question: str, additional_data: dict = {}) -> dict:
        query = """
        query GetFunction($question: String!, $staticFunctionArguments: [StaticFunctionArgument]) {
            get_and_instantiate_function(question: $question, static_function_arguments: $staticFunctionArguments) {
                ... on SQLFunction {
                function_name
                description
                post_processing_code_template
                instantiated_post_processing_code
                arguments {
                    name
                    description
                    general_type
                    is_user_editable
                    instantiated_value
                    available_values
                }
                sql_template
                instantiated_sql
            }
            }
        }
        """
        static_function_arguments = [{"name": key, "value": str(value)} for key, value in additional_data.items()]
        variables = {"question": question, "staticFunctionArguments": static_function_arguments}
        response = requests.post(self._graphql_endpoint, headers=self._graphql_headers, json={'query': query, 'variables': variables})
        response_json = response.json()
        if response.status_code == 200 and 'data' in response_json and 'get_and_instantiate_function' in response_json['data']:
            self.log(response_json['data']['get_and_instantiate_function'])
            resp = response_json['data']['get_and_instantiate_function']

            print(resp)

            return resp
        else:
            raise Exception(f"Query failed to run by returning code of {response.status_code}. {response.text}")

    def create_function(self, question: str, sql: str, plotly_code: str, **kwargs) -> dict:
        query = """
        mutation CreateFunction($question: String!, $sql: String!, $plotly_code: String!) {
            generate_and_create_sql_function(question: $question, sql: $sql, post_processing_code: $plotly_code) {
                function_name
                description
                arguments {
                    name
                    description
                    general_type
                    is_user_editable
                }
                sql_template
                post_processing_code_template
            }
        }
        """
        variables = {"question": question, "sql": sql, "plotly_code": plotly_code}
        response = requests.post(self._graphql_endpoint, headers=self._graphql_headers, json={'query': query, 'variables': variables})
        response_json = response.json()
        if response.status_code == 200 and 'data' in response_json and response_json['data'] is not None and 'generate_and_create_sql_function' in response_json['data']:
            resp = response_json['data']['generate_and_create_sql_function']

            print(resp)

            return resp
        else:
            raise Exception(f"Query failed to run by returning code of {response.status_code}. {response.text}")

    def update_function(self, old_function_name: str, updated_function: dict) -> bool:
        """
        Update an existing SQL function based on the provided parameters.

        Args:
            old_function_name (str): The current name of the function to be updated.
            updated_function (dict): A dictionary containing the updated function details. Expected keys:
                - 'function_name': The new name of the function.
                - 'description': The new description of the function.
                - 'arguments': A list of dictionaries describing the function arguments.
                - 'sql_template': The new SQL template for the function.
                - 'post_processing_code_template': The new post-processing code template.

        Returns:
            bool: True if the function was successfully updated, False otherwise.
        """
        mutation = """
        mutation UpdateSQLFunction($input: SQLFunctionUpdate!) {
            update_sql_function(input: $input)
        }
        """

        SQLFunctionUpdate = {
            'function_name', 'description', 'arguments', 'sql_template', 'post_processing_code_template'
        }

        # Define the expected keys for each argument in the arguments list
        ArgumentKeys = {'name', 'general_type', 'description', 'is_user_editable', 'available_values'}

        # Function to validate and transform arguments
        def validate_arguments(args):
            return [
                {key: arg[key] for key in arg if key in ArgumentKeys}
                for arg in args
            ]

        # Keep only the keys that conform to the SQLFunctionUpdate GraphQL input type
        updated_function = {key: value for key, value in updated_function.items() if key in SQLFunctionUpdate}

        # Special handling for 'arguments' to ensure they conform to the spec
        if 'arguments' in updated_function:
            updated_function['arguments'] = validate_arguments(updated_function['arguments'])

        variables = {
            "input": {
                "old_function_name": old_function_name,
                **updated_function
            }
        }

        print("variables", variables)

        response = requests.post(self._graphql_endpoint, headers=self._graphql_headers, json={'query': mutation, 'variables': variables})
        response_json = response.json()
        if response.status_code == 200 and 'data' in response_json and response_json['data'] is not None and 'update_sql_function' in response_json['data']:
            return response_json['data']['update_sql_function']
        else:
            raise Exception(f"Mutation failed to run by returning code of {response.status_code}. {response.text}")

    def delete_function(self, function_name: str) -> bool:
        mutation = """
        mutation DeleteSQLFunction($function_name: String!) {
            delete_sql_function(function_name: $function_name)
        }
        """
        variables = {"function_name": function_name}
        response = requests.post(self._graphql_endpoint, headers=self._graphql_headers, json={'query': mutation, 'variables': variables})
        response_json = response.json()
        if response.status_code == 200 and 'data' in response_json and response_json['data'] is not None and 'delete_sql_function' in response_json['data']:
            return response_json['data']['delete_sql_function']
        else:
            raise Exception(f"Mutation failed to run by returning code of {response.status_code}. {response.text}")

    def create_model(self, model: str, **kwargs) -> bool:
        """
        **Example:**
        ```python
        success = vn.create_model("my_model")
        ```
        Create a new model.

        Args:
            model (str): The name of the model to create.

        Returns:
            bool: True if the model was created, False otherwise.
        """
        model = sanitize_model_name(model)
        params = [NewOrganization(org_name=model, db_type="")]

        d = self._rpc_call(method="create_org", params=params)

        if "result" not in d:
            return False

        status = Status(**d["result"])

        return status.success

    def get_models(self) -> list:
        """
        **Example:**
        ```python
        models = vn.get_models()
        ```

        List the models that belong to the user.

        Returns:
            List[str]: A list of model names.
        """
        d = self._rpc_call(method="list_my_models", params=[])

        if "result" not in d:
            return []

        orgs = OrganizationList(**d["result"])

        return orgs.organizations

    def generate_embedding(self, data: str, **kwargs) -> list[float]:
        # This is done server-side
        pass

    def add_question_sql(self, question: str, sql: str, **kwargs) -> str:
        if "tag" in kwargs:
            tag = kwargs["tag"]
        else:
            tag = "Manually Trained"

        params = [QuestionSQLPair(question=question, sql=sql, tag=tag)]

        d = self._rpc_call(method="add_sql", params=params)

        if "result" not in d:
            raise Exception("Error adding question and SQL pair", d)

        status = StatusWithId(**d["result"])

        return status.id

    def add_ddl(self, ddl: str, **kwargs) -> str:
        params = [StringData(data=ddl)]

        d = self._rpc_call(method="add_ddl", params=params)

        if "result" not in d:
            raise Exception("Error adding DDL", d)

        status = StatusWithId(**d["result"])

        return status.id

    def add_documentation(self, documentation: str, **kwargs) -> str:
        params = [StringData(data=documentation)]

        d = self._rpc_call(method="add_documentation", params=params)

        if "result" not in d:
            raise Exception("Error adding documentation", d)

        status = StatusWithId(**d["result"])

        return status.id

    def get_training_data(self, **kwargs) -> pd.DataFrame:
        params = []

        d = self._rpc_call(method="get_training_data", params=params)

        if "result" not in d:
            return None

        # Load the result into a dataclass
        training_data = DataFrameJSON(**d["result"])

        df = pd.read_json(StringIO(training_data.data))

        return df

    def remove_training_data(self, id: str, **kwargs) -> bool:
        params = [StringData(data=id)]

        d = self._rpc_call(method="remove_training_data", params=params)

        if "result" not in d:
            raise Exception("Error removing training data")

        status = Status(**d["result"])

        if not status.success:
            raise Exception(f"Error removing training data: {status.message}")

        return status.success

    def get_related_training_data_cached(self, question: str) -> TrainingData:
        params = [Question(question=question)]

        d = self._rpc_call(method="get_related_training_data", params=params)

        if "result" not in d:
            return None

        # Load the result into a dataclass
        training_data = TrainingData(**d["result"])

        self.related_training_data[question] = training_data

        return training_data

    def get_similar_question_sql(self, question: str, **kwargs) -> list:
        if question in self.related_training_data:
            training_data = self.related_training_data[question]
        else:
            training_data = self.get_related_training_data_cached(question)

        return training_data.questions

    def get_related_ddl(self, question: str, **kwargs) -> list:
        if question in self.related_training_data:
            training_data = self.related_training_data[question]
        else:
            training_data = self.get_related_training_data_cached(question)

        return training_data.ddl

    def get_related_documentation(self, question: str, **kwargs) -> list:
        if question in self.related_training_data:
            training_data = self.related_training_data[question]
        else:
            training_data = self.get_related_training_data_cached(question)

        return training_data.documentation
